"""
Recip -- Sage package for using Shimura's reciprocity law

See the file README.txt for version information and instructions.

#*****************************************************************************
# Copyright (C) 2010 -- 2012 Marco Streng <marco.streng@gmail.com>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#*****************************************************************************

this file:
denominators.sage

This file computes proven Igusa class polynomials.


"""


def class_polynomials(K, factor=False, prec=None, D=None,
                      verbose=False):
    """
    Returns the Igusa class polynomials of K
    (proven!).
    
    EXAMPLES:
    
    We prove compute proven class polynomials for some of the entries of
    van Wamelen's table. With these class polynomials, it is trivial to
    check correctness of the table entries.::
    
        sage: load("recip.sage")
        sage: K = CM_Field([5,5,5])
        sage: class_polynomials(K) # long time
        [x, 0, 0]

        sage: class_polynomials(CM_Field((x^2+2)^2-2), verbose=True) # long time
        starting with precision 50
        output has 309.861183858049 too few bits of precision
        increasing precision to 364
        output has 2.24686161486127 more bits of precision than needed
        Denominator: 1 out of 2^274 * 31^8
        [x + 7290, 437400, 2952450000]

        sage: class_polynomials(CM_Field((x^2+13)^2-2^2*13), verbose=True) # long time
        starting with precision 50
        output has 11.3211441914114 too few bits of precision
        increasing precision to 66
        output has 5.23757254036671 more bits of precision than needed
        Denominator: 1 out of 2^14
        [x + 7840, 102400, -204800000]

        sage: class_polynomials(CM_Field((x^2+5)^2-5)) # not tested (too hard)
        starting with precision 4116

        sage: class_polynomials(CM_Field((x^2+65)^2-26^2*5)) # not tested (too hard)

        sage: bruinier_yang_applies(CM_Field((x^2+29)^2-4*29))
        True
        sage: class_polynomials(CM_Field((x^2+29)^2-2^2*29), verbose=True) # long time
        starting with precision 50
        output has 37.1180995525538 too few bits of precision
        increasing precision to 92
        output has 4.77580985893192 more bits of precision than needed
        Denominator: 1 out of 2^14 * 5^4
        [x + 2589408, 131383296, -60466176000000]
        
        sage: bruinier_yang_applies(CM_Field((x^2+61)^2-6^2*61))
        True
        sage: class_polynomials(CM_Field((x^2+61)^2-6^2*61), verbose=True) # long time
        starting with precision 50
        output has 55.7922056566579 too few bits of precision
        increasing precision to 110
        output has 3.94423036431886 more bits of precision than needed
        Denominator: 3^3 * 41^4 out of 2^14 * 3^8 * 5^4 * 41^4
        [x - 88833024/1681, -14055214415872/76295547, 9663676416000000/2825761]

    """
    if factor:
        raise NotImplementedError, "factor=True not yet implemented"
    Zs = list(K.period_matrices_iter())
    h = igusa_modular_forms()
    if D is None:
        D = denominator_bound(K)
    # compute the period matrices and denominators here
    if prec is None:
        # prec = ZZ(floor(40 + log(D)/log(2)))
        # or maybe some more sensible guess?
        prec = 50
        if verbose:
            print "starting with precision %s" % prec
            sys.stdout.flush()

    while True:
        hom_values = [[f(Z, prec=prec, interval=True) for f in h] for Z in Zs] 
        abs_values = [igusa_modular_forms_to_absolute(Is) for Is in hom_values]
        abs_values = [[z[k] for z in abs_values] for k in range(3)]
        x = hom_values[0][0].parent()['x'].gen()
        pols = [prod([x-i1 for i1 in abs_values[0]])]
        pols = pols + [short_interpolation(abs_values[0], abs_values[k]) for k in [1,2]]
        pols = [[x.real() for x in (D*p).list()] for p in pols]
        # These polynomials are approximations of integer polynomials.
        err = max(flatten([[x.absolute_diameter() for x in p] for p in pols]))
        err = (log(err)/log(2)).n()
        if err < 0:
            if verbose:
                print "output has %s more bits of precision than needed" % -err
            # we found the polynomial!
            try:
                pols = [[c.unique_integer() for c in p] for p in pols]
            except ValueError as e:
                raise RuntimeError, "Incorrect denominator or bug, if you did "\
                                    "not specify a denominator, please report "\
                                    "%s" % e
            ret = [QQ['x'](p)/D for p in pols]
            if verbose:
                print "Denominator: %s out of %s" % \
                      (lcm([p.denominator() for p in ret]).factor(), D.factor())
            return ret
        else:
            if verbose:
                print "output has %s too few bits of precision" % err
            # intervals are still pretty large, so we increase precision
            prec = prec + ZZ(floor(err)) + 5
            if verbose:
                print "increasing precision to %s" % prec
                sys.stdout.flush()

