"""
Recip -- Sage package for using Shimura's reciprocity law

See the file README.txt for version information and instructions.

#*****************************************************************************
# Copyright (C) 2010 -- 2012 Marco Streng <marco.streng@gmail.com>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#*****************************************************************************

this file:
denominators.sage

This file contains implementations of the denominator bounds of
Goren-Lauter and Bruinier-Yang.

References for this file:

[Yang] -    ARITHMETIC INTERSECTION ON A HILBERT MODULAR SURFACE AND THE
            FALTINGS HEIGHT -- Tonghai Yang
            http://www.math.wisc.edu/~thyang/general4L.pdf

[GJLSVW] -  IGUSA CLASS POLYNOMIALS, EMBEDDINGS OF QUARTIC CM FIELDS, AND
            ARITHMETIC INTERSECTION THEORY -- HELEN GRUNDMAN, JENNIFER
            JOHNSON-LEUNG, KRISTIN LAUTER, ADRIANA SALERNO, BIANCA VIRAY,
            ERIKA WITTENBORN
            http://arxiv.org/abs/1006.0208
"""


def denominator_bound(K, c=2**14, k=2, d=None, Phi=None, bound='default', check=True):
    """
    If ``Phi`` is None, returns a positive integer `D`
    such that `D N(f h_{10}^{-k}(Z))`
    is integral whenever `Z` has CM by `O_K`,
    `c*f` is a modular
    form of weight `k` and level `1` with integral
    Fourier coefficients, and
    `d` (if K/QQ is Galois) and `2d` (if K/QQ is non-Galois)
    is the degree of the algebraic number 
    `f h_{10}^{-k}(Z)`.
    
    For the choice of invariants in the author's thesis, take k=2 and c=2^14.
    These are also the defaults.
        
    If ``Phi`` is a CM-type, returns a number `D`
    in the real quadratic subfield of the reflex field
    of ``Phi`` such that
    that `D f h_{10}^{-k}(Z)`
    is integral whenever `Z` has CM by `O_K`
    of type ``Phi`` and `c*f` is a modular
    form of weight `k` and level `1` with integral
    Fourier coefficients.
    
    EXAMPLE:
    
    The following example shows how horribly unusable the Goren-Lauter bounds
    are in practice. The additional number of digits (about 2000) needed to
    compute the class polynomial with certainty using the algorithm of the
    author's thesis is huge even though it is also easy to see that actually
    the class polynomials are all equal to X, so the algorithm as given is not
    practical. The importance of the algorithm is in proving the asymptotic
    running time::
    
        sage: load("recip.sage")
        sage: gl = denominator_bound(CM_Field([5,5,5]), bound='gl')
        sage: gl.factor()
        2^14 * 5^36 * 19^20 * 29^18 * 59^14 * 79^14 * 89^14 * 109^12 * 139^12 * 149^12 * 179^12 * 199^12 * 229^12 * 239^12 * 269^10 * 349^10 * 359^10 * 379^10 * 389^10 * 409^10 * 419^10 * 439^10 * 449^10 * 479^10 * 499^10 * 509^10 * 569^10 * 599^10 * 619^10 * 659^10 * 709^10 * 719^10 * 739^10 * 769^10 * 809^10 * 829^10 * 839^10 * 859^10 * 919^10 * 929^10 * 1009^8 * 1019^8 * 1039^8 * 1049^8 * 1069^8 * 1109^8 * 1129^8 * 1229^8 * 1249^8 * 1259^8 * 1279^8 * 1289^8 * 1319^8 * 1399^8 * 1409^8 * 1429^8 * 1439^8 * 1459^8 * 1489^8 * 1499^8 * 1549^8 * 1559^8 * 1579^8 * 1609^8 * 1619^8 * 1669^8 * 1699^8 * 1709^8 * 1759^8 * 1789^8 * 1879^8 * 1889^8 * 1949^8 * 1979^8 * 1999^8
        sage: (log(gl)/log(10)).n()
        1947.25745024934
        sage: (log(gl)/log(2)).n()
        6468.64923196203
    
    Fortunately the Bruinier-Yang bounds apply::
    
        sage: denominator_bound(CM_Field([5,5,5])).factor()
        2^14
    
    Next, we try a different field::
                        
        sage: K = CM_Field((x^2+2)^2-2,'a')
        sage: bruinier_yang_applies(K)
        False
        sage: gl = denominator_bound(K)
        sage: (log(gl)/log(2)).n()
        313.633570483095
        sage: Z = K.one_period_matrix(); Z
        Period Matrix
        [0.?e-15 + 1.08239220029240?*I 0.41421356237310? + 0.?e-14*I]
        [            0.41421356237310?           1.53073372946036?*I]
        sage: h = igusa_modular_forms()
        sage: v = [f(Z, prec=100, interval=True) for f in h]; v
        [5.14511662180777441738853320? + 0.?e-28*I, 1.5657740361761366552190052? + 0.?e-27*I, -0.00110508779417351004090108656? + 0.?e-30*I, 0.0201781755750066641900463261? + 0.?e-29*I]
        sage: w = igusa_modular_forms_to_absolute(v); w
        [-7290.0000000000000000000000? + 0.?e-23*I, 437400.00000000000000000000? + 0.?e-21*I, 2.9524500000000000000000000?e9 + 0.?e-17*I]
        sage: [(log((gl*y).real().absolute_diameter())/log(2)).n() for y in w]
        [241.441063762631, 247.704281928552, 260.403012023575]

    These numbers are very recognizable, but we need 346 more bits of output
    precision to prove they are correct.::
 
        sage: v = [f(Z, prec=450, interval=True) for f in h]
        sage: w = igusa_modular_forms_to_absolute(v)
        sage: x = [gl*y.real() for y in w]
        sage: [(log(y.absolute_diameter())/log(2)).n() for y in x]
        [-106.982082440410, -100.796499047639, -88.0728007502253]

    So now we have more than enough precision and can simply round::
    
        sage: z = [y.real().unique_integer() for y in x]
        sage: invs = [y/gl for y in z]; invs
        [-7290, 437400, 2952450000]
    
    These numbers are proven correct, so let's prove the corresponding entry
    of van Wamelen's table with it. First, we double-check that the curve
    in the table corresponds with the invariants given in the table::
    
        sage: x = QQ['x'].gen()
        sage: f = -x^5+3*x^4+2*x^3-6*x^2-3*x+1
        sage: [I2,I4,I6,I10] = igusa_clebsch_invariants(f)
        sage: (I2^5/I10) == 2^7*3^15
        True
        sage: I2^3*I4/I10 == 2^5*3^11*5
        True
        sage: I2^2*I6/I10 == 2^4*3^9*31
        True
    
    Next, we show that these invariants coincide with the value we just proved.
    This seems to be the first complete proof of correctness for any entry
    of van Wamelen's table, other than the classical and easy QQ(zeta_5).
    It was already proven by van Wamelen that the curve `y^2 = f` has CM
    by an order of `K`, and a sketch was given of how to extend these results
    to show the order is the maximal order. Now we have a complete proof (by
    different means).::
    
        sage: I6pr = (I2*I4-3*I6)/2
        sage: invs == [I4*I6pr/I10, I4^2*I2/I10, I4^5/I10^2]
        True

    Check that a bug with incorrect embeddings is fixed::
    
        sage: K = CM_Field([5,26,149])
        sage: denominator_bound(K).factor()
        2^14 * 5^4 * 7^4

    """
    if d is None:
        d = K.class_number()
        if Phi is None and not K.is_galois():
            d = d * 2
    if not Phi is None:
        raise NotImplementedError
    if bound == 'default':
        if bruinier_yang_applies(K):
            bound = 'by'
            check = False # as the check has already been done
        else:
            bound = 'gl'
    if bound == 'gl':
        b = goren_lauter_bound(K, Phi=Phi)**d
    elif bound == 'by':
        b = bruinier_yang_bound(K, Phi=Phi, check=check)
    else:
        raise ValueError, "Unkown bound: %s" % bound
    gal = K.is_galois()
    if not gal:
        c = sqrt(c)
    ret = c**d * b**k
    if ret in QQ:
        return QQ(ret)
    else:
        # There may be some cases where the output is the square root of an
        # integer. We take the largest integer "dividing" that square root:
        return QQ([p**ZZ(floor(e/2)) for (p,e) in QQ(ret**2).factor()])


def bruinier_yang_applies(K, proof=True, reason=False):
    """
    Return ``True`` if and only if the Bruinier-Yang results are proven for
    the input quartic CM field in Yang's preprint.
    
    With ``proof=False`` returns ``True`` only if the Bruinier-Yang results are
    formulated and conjectured for the input quartic CM field
    as in Yang's preprint.
    
    With ``reason=True`` also return a string with an explanation.
    
    EXAMPLES::
    
        sage: load("recip.sage")
        sage: bruinier_yang_applies(CM_Field([5,5,5]))
        True
    """
    #1Assumption from page 2: D is a prime that is 1 mod 4
    # Notation: F = QQ(sqrt(D))) is real quadratic subfield of K
    #           Delta is such that K=F(sqrt(Delta))
    #           Delta' is complex conjugate of Delta
    #           Dtilde is relative norm of Delta
    #           Ktilde is reflex field
    #           Ftilde is real quadratic subfield of reflex field
    #2Assumption from page 2: Dtilde is the discriminant of Ftilde
    #           (this is presented as notation, but does a Delta such that this
    #            holds always exist? also, Delta appears later on, and
    #            if we let Dtilde be the discriminant, then
    #            Dtilde * m^2 = Delta*Delta' for a positive integer m,
    #            but m does not have to be square in F. Also, it is part of the
    #            assumption a few lines lower.)
    #3Assumption from page 3: O_K is free over O_F
    #4Assumption: O_K can actually be written as O_F + O_F((w+sqrt(Delta))/2)
    #           (Yang's phrasing suggeststhat this is not stronger than
    #            "O_K is free over O_F", still, this seems to really be an
    #            assumption. Moreover, this Delta is the same as the one in
    #            Dtilde = Delta*Delta')
    #5Assumption from Theorem 1.2: Dtilde = Delta*Delta' is 1 modulo 4 and prime
    #
    delta0 = K.real_field().discriminant()
    if not is_prime(delta0):
        # 1 is not satisfied
        if reason:
            return False, "Bruinier-Yang formula not formulated as K0=QQ(sqrt(%s)) does not have prime discriminant" % delta0
        return False
    # now we know 1 is satisfied

    # Let A = {x-xbar : x in OK}, which generates the relative different diff.
    # If the hypothesis is satisfied, then 
    # A = sqrt(Delta)*OF, so A*OK=diff.
    Krel = K.relative_field()
    diff = Krel.relative_different()
    if not diff.is_principal():
        return False
    eta = diff.gens_reduced()[0]
    eps = K.unit_group().gens()[1]
    assert eps.multiplicative_order() == oo
    eps = Krel.structure()[1](eps) # move it to the relative field
    
    if eta^2 in Krel.base_field():
        pass
    elif eta^2*eps^2 in Krel.base_field():
        eta = eta*eps
    else:
        # 4 is not satisfied
        if reason:
            return False, "Bruinier-Yang formula not formulated as OK is not monogenic of the right form over OF"
        return False
    Delta = K.real_field()(eta^2)
    
    # If we reach this point, then we have found Delta in F with K=F(sqrt(Delta))
    # and Delta*Delta' is prime, hence is the discriminant of Ftilde.
    # So 1, 2, 5 are satisfied. If 4 is satisfied, then so is 3.
    # Next, considering the minimal polynomial of the algebraic number
    # (w+sqrt(Delta))/2,
    # we get
    # ((w+sqrt(Delta))/2)^2 = w^2/4 + w*sqrt(Delta)/2 + Delta/4
    # = w * (w+sqrt(Delta))/2 - w^2/4 + Delta/4, so 
    # Delta is w^2 modulo 4 if and only if the number is an algebraic integer.
    OF = K.real_field().maximal_order()
    for w in (4*OF).residues():
        if (4*OF).divides(Delta-w^2):
            # 4 (and hence 3) is satisfied
            break
    else: # this "else" belongs to the "for"
        # 4 is not satisfied. Indeed, suppose 4 is true.
        # Delta in the Sage code above is uniquely defined, up to squares of units
        # in K. So up to squares of units in K, it is the Delta in 4.
        # Now squares of units in K, if they are in F, are also squares of units
        # in F (Lemma II.3.3 of my thesis).
        # So Delta is correct up to squares of units.
        if reason:
            return False, "Bruinier-Yang formula not formulated as OK is not monogenic of the right form over OF"
    
    # Now we know 1, 3, 4 are satisfied.
    
    Dtilde = QuadraticField(K.DAB()[2]).discriminant()
    if Delta.norm() != Dtilde:
        # We know 1, 3, 4, so Delta generates the relative discriminant
        # and is totally negative, 
        # hence is defined up to totally positive units. Its norm is therefore
        # defined by 1, 3, 4. So we now know that there is no Delta
        # at all satisfying 1 -- 4.
         
        # If Delta.norm() is prime, then we never reach this line.
        assert not ZZ(Delta.norm()).is_prime()
        assert (Delta.norm() / Dtilde).sqrt() in QQ
        
        if reason:
            return False, "Bruinier-Yang formula not formulated, at least not if both Dtilde's are equal (Delta*Delta' and disc(Ftilde)) and both Delta's are equal (in the generator of OK/OF and in Delta*Delta'=Dtilde)"
        return False
    
    # Now we know 1 -- 4 are satisfied.
    
    if not proof:
        if reason:
            return True, "Bruinier-Yang's conjecture has been stated as a conjecture in this case."
        return True
    
    if not is_prime(ZZ(K.discriminant()/delta0**2)):
        if reason:
            return False, "Bruinier-Yang formula conjectured, but not proven as the relative discriminant (%s) is not prime" % ZZ(K.discriminant()/delta0**2)
        return False
        
    if reason:
        return True, "Bruinier-Yang's conjecture has been proven in this case."
    return True


def bruinier_yang_bound(K, Phi=None, check=True, proof=True):
    """
    Returns the Bruinier-Yang bounds. Assumes
    that the input satisfies the hypotheses of the
    Bruinier-Yang conjecture. Bounds are proven
    only if ``bruinier_yang_applies`` returns ``True``.
    
    Phi is not used.
    
    EXAMPLES::
    
        sage: load("recip.sage")
        sage: bruinier_yang_bound(CM_Field([5,5,5]))
        1
    
    Let's check for a few fields how sharp these denominators are. First,
    all elements of van Wamelen's table where the conjecture has been
    proven.::
    
        sage: lst = wamelen_dab_list()
        sage: lst = [(DAB, CM_Field(DAB)) for DAB in lst]
        sage: lst = [(DAB, K, bruinier_yang_applies(K, proof=True)) for (DAB, K) in lst if bruinier_yang_applies(K, proof=False)]
        sage: len(lst)
        6
        sage: [(DAB, b) for (DAB, K, b) in lst]
        [([5, 5, 5], True), ([13, 13, 13], True), ([29, 29, 29], True), ([37, 37, 333], True), ([53, 53, 53], True), ([61, 61, 549], True)]
        sage: lst = [(K, denominator_bound(K), class_polynomials(K)) for (DAB, K, b) in lst if b] # long time
        sage: len(lst) # long time
        6
        sage: [(K.DAB(), den.factor(), lcm([c.denominator() for c in flatten([list(p) for p in pols])]).factor()) for (K,den,pols) in lst] # long time
        [([5, 5, 5], 2^14, 1), ([13, 13, 13], 2^14, 1), ([29, 29, 29], 2^14 * 5^4, 1), ([37, 37, 333], 2^14 * 3^4 * 11^4, 11^2), ([53, 53, 53], 2^14 * 17^4 * 29^4, 17^2 * 29^4), ([61, 61, 549], 2^14 * 3^8 * 5^4 * 41^4, 3^3 * 41^4)]

    So we see that den contains too many 2's, 5's, 11's in the first cases, but
    den is optimal wrt 29 and 41 in the last two cases. The 2, 3, 5, 11, 17
    could be cancellation, but is still a bit fishy. Anyway, here is the list
    of 6 triples of class polynomials that we just proved::
    
        sage: [(K.DAB(), pols) for (K,den,pols) in lst] # long time
        [([5, 5, 5], [x, 0, 0]), ([13, 13, 13], [x + 7840, 102400, -204800000]), ([29, 29, 29], [x + 2589408, 131383296, -60466176000000]), ([37, 37, 333], [x + 505440, -810393600/121, 1642291200000]), ([53, 53, 53], [x + 16982274516480/14297, 30036725005359513600/204404209, -16680511371779820748800000/707281]), ([61, 61, 549], [x - 88833024/1681, -14055214415872/76295547, 9663676416000000/2825761])]

    Here is the final curve in the author's preprint with Florian Bouyer::
    
        sage: K = CM_Field([389, 37, 245])
        sage: class_polynomials(K, verbose=True)
        starting with precision 50
        output has 117.335551734613 too few bits of precision
        increasing precision to 172
        output has 4.33681435945913 more bits of precision than needed
        Denominator: 11^2 * 19^6 * 29^4 out of 2^14 * 5^4 * 11^8 * 19^8 * 29^4
        [x^2 - 502951680/303601*x - 215457915617280/5768419, 102098407935836160/11153001631321*x + 1497833452550013478502400/4026233588906881, 2213545771008000000/92173567201*x - 36646804488714190848000000/33274657759561]
    """
    if check:
        b, s = bruinier_yang_applies(K, proof=proof, reason=True)
        if not b:
            raise ValueError, s
        if get_recip_verbose():
            print s
    if get_recip_verbose():
        print "Computing Bruinier-Yang's bound"

    # We need to take the number of roots of unity into account.
    # The logarithm of what we have below is a factor #OK^*^tors / 2 too small.
    # This factor is 5 if K=QQ(zeta_5) and 1 otherwise.
    # We can ignore the 1, and in the case of QQ(zeta_5), our output is 1
    # anyway, so raising to the power 5 does not change that.
    # Therefore, we can safely ignore the factor W_K/2 in [Yang].
    
    # We start with some standard fields and discriminants.
    Phi = K.CM_types()[0]
    Kr = Phi.reflex_field()                 # denoted Ktilde in Yang
    Kr_rel = Kr.relative_field()            # denoted Ktilde in Yang
#    K0r = Phi.reflex_field().real_field()   # denoted Ftilde in Yang
    K0 = K.real_field()                     # denoted F in Yang
    D = K0.disc()                           #         D
    delta1 = K.disc()/D**2                  # this equals Dtilde by assumptions
    d = Kr_rel.relative_discriminant()      # denoted d_{Ktilde/Ftilde} in Yang
    K0r = Kr_rel.base_field()               # denoted Ftilde in Yang
    Dtilde = K0r.disc()                     #         Dtilde
    B = sqrt(K0r(Dtilde))                   # in Yang: sqrt(Dtilde) (gen. of F)
    O0r = K0r.maximal_order()
    Kr_rel_to_Kr = Kr_rel.structure()[0]
    def cu(u):
        """
        Given u, which is one of the numbers t in F in the sum in (1.3),
        return exp(sum_{P prime of Ftilde | p} B_t(P) log(p)).
        In other words, if we sum over S, we get exp(2*RHS of (1.11)).
        """
        if get_recip_verbose()>1:
            print "Computing contribution of\nt = %s\n  = d^-1 * (%s)\n  = (k+(D-n^2)/4*sqrt(Dtilde))/(2D),\nwhere k=%s and (D-n^2)/4=%s and n=%s" % (u, u*d.gens_reduced()[0], (t*D).trace(), (D*t*2-(D*t).trace())/B, sqrt(D-(D*t*2-(D*t).trace())/B*4))
        f = (u*d).factor()
        def splittype(p):
            """
            Given a prime P of Ftilde, returns
            0 if P splits in Ktilde/Ftilde
            1 if P is inert in Ktilde/Ftilde
            2 if P ramifies in Ktilde/Ftilde
            """
            t = Kr_rel_to_Kr(p).factor()
            if len(t) == 2:
                return 0
            return t[0][1]
            
        # For each prime P of Ftilde, take (P, val_P(u*d), splittype)
        # if val_P(u*d) is non-zero:
        f2 = [(p[0], p[1], splittype(p[0])) for p in f]
        
        if get_recip_verbose()>1:
            print "Primes over t*d: (given by [p,e,f] in Ftilde/QQ)"
            for p in f2:
                if p[2] == 0:
                    s = "  Split in Ktilde/Ftilde:    "
                elif p[1] == 1:
                    s = "  Inert in Ktilde/Ftilde:    "
                else:
                    s = "  Ramified in Ktilde/Ftilde: "
                print s + "[%s,%s,%s] to power %s" % (p[0].gens_two()[0], p[0].ramification_index(), p[0].residue_class_degree(), p[1])
        
        def prerho(k, t):
            """
            Given an integer k and the output of splittingtype t for a prime
            P of Ftilde, returns the number of ideals of Ktilde of relative
            norm P^k, i.e., returns rho(P^k) as below (1.4).
            """
            if k < 0:  # not an integral ideal
                return 0
            if k == 0: # the unit ideal
                return 1
            if t == 2: # ramified
                return 1
            if t == 0: # split
                return k+1
            # inert:
            if is_even(k):
                return 1
            return 0

        def rho(k):
            """
            Given the index k of an entry (P,v,t)=f2[k] of f2, returns
            the number of ideals of Ktilde of relative norm u*d*P^-1,
            i.e., returns rho(u*d*P^-1) as below (1.4).
            
            Actually, it returns this number only if P is not split. If P
            is split, then because of (1.4) the number 0 is returned instead.
            """
            if f2[k][2] == 0:
                return 0
            return prod((prerho(f2[l][1], f2[l][2]) for l in range(len(f2)) if k != l))*prerho(f2[k][1]-1,f2[k][2])

        # ord_P(t) = f2[k][1] - d.valuation(f2[k][0])
        ret = prod(
                   f2[k][0].norm() ** ((f2[k][1] - d.valuation(f2[k][0]) + 1) * (rho(k))) for k in range(len(f2))
                  )
        if get_recip_verbose() > 1:
            print "Contribution of t is %s" % ret
        return ret
        
    # Some notation changes within Yang:
    # n of page 5 is denoted n by us
    # n of page 3 is denoted k by us
    # m of page 5 is 1 in our case
    # m of pages 2 and 3 is denoted (D-n^2)/4 by us
    
    # a is the set of pairs (k,n) where n is an odd integer and k is an
    #                             integer with k^2 <= (D-n^2)^2 Dtilde /4    
    a = [[(k,n) for k in range(-floor((D-n^2)*sqrt(Dtilde)/4),floor((D-n^2)*sqrt(Dtilde)/4)+1)] for n in range(1,floor(sqrt(D))+1,2)]

    # b is the set of elements t = k+m*sqrt(Dtilde))/(2D) in F, where
    # (k,n) is in a, and m = (D-n^2)/4 is the subscript of "b".
    b = [(4*k+(D-n^2)*B)/(8*D) for (k,n) in flatten(a, max_level=1)]

    # Equation (1.3) says we should restrict to t in d_{Ktilde/Ftilde}^-1 = d.
    # So we define S to be the subset of b satisfying this condition.
    # The further condition |n|<m sqrt(Dtilde) in (1.3) is in our notation
    # |k| < (D-n^2)/4 * Dtilde, which is exactly the restriction we already
    # have.
    # So S is the set of t in equation (1.3), but where m also ranges
    # over all the numbers m = (D-n^2)/4 of page 5.
    S = [u for u in b if u in d^-1]
    
    #maxnrofls = max([len([elt for elt in S if elt[1] == n]) for n in range(1,floor(sqrt(D))+1,2)]+[0])
    # sizeofS = len(S)
    # maxsizeforoneeltofS = max([log((cu(u))+0.0) for u in S]+[0])
    ret = prod([cu(t) for t in S])
    # By Remark 7.1 of [GJLSVW], if K/QQ is Galois, then we have counted the
    # denominator twice, so:
    if K.is_galois():
        if ret.is_square():
            ret = ret.sqrt()
        else:
            print "K is Galois, but the Bruinier-Yang arithmetic intersection number is not the logarithm of a square. Interesting..."
            ret = ret.sqrt()
            # TODO: I would like to round down. Is that ok? Not sure, so I don't.
            #f = ret.factor()
            #ret = [p**ZZ(floor(e/2)) for (p,e) in f]
    return ret


def goren_lauter_bound(K, Phi=None):
    """
    If ``Phi`` is None, returns a positive integer `D`
    such that `c^l D^{dk} N(f h_{10}^{-k}(Z))`
    is integral whenever `Z` has CM by `O_K`,
    `f` is a modular
    form of weight `k` and level `1` with integral
    Fourier coefficients, and
    `d` (if K/QQ is Galois) and `2d` (if K/QQ is non-Galois)
    is the degree of the algebraic number 
    `f h_{10}^{-k}(Z)`.
    
    Here l is the degree and is l=d if K/QQ is Galois and l=2*d otherwise.
    
    For the choice of invariants in the author's thesis, take k=2 and note that
    2^14*i_n satisfies the hypotheses.
        
    If ``Phi`` is a CM-type, returns a number `D`
    in the real quadratic subfield of the reflex field
    of ``Phi`` such that
    that `D^k f h_{10}^{-k}(Z)`
    is integral whenever `Z` has CM by `O_K`
    of type ``Phi`` and `f` is a modular
    form of weight `k` and level `1` with integral
    Fourier coefficients.
    """
    if not Phi is None:
        raise NotImplementedError
    gal = K.is_galois()
    prime_powers = []
    [D,A,B] = DAB_to_minimal(K.DAB())
    a = A/2
    d = ZZ(D)
    if d%4 == 0:
        d = ZZ(d/4)
    b = sqrt((A**2-4*B)/d)/2
    k = ZZ(floor(min([a.valuation(2), b.valuation(2)])/2))
    a = ZZ(a/2**(2*k))
    b = ZZ(b/2**(2*k))
    primes = prime_range(4*d*a**2)
    prime_powers = []
    disc = K.discriminant()
    for p in primes:
        # this is the place to test whether the splitting of p is right
        if gal: # for some splitting types we can take 1 even if gal is False
            if not (disc % p):
                galfactor = 1
            elif len(K.ideal(p).factor()) == 2:
                galfactor = 1
            else:
                galfactor = 0
        else:
            if Phi is None:
                Kr = K.CM_types()[0].reflex_field()
            else:
                Kr = Phi.reflex_field()
            if len(K.ideal(p).factor()) == 2:
                f = Kr.ideal(p).factor()
                if len(f) == 2:
                    galfactor = 2
                elif len(f) == 3:
                    galfactor = 1
                else:
                    galfactor = 0
            else:
                galfactor = 0
            galfactor = 2
        base = p**galfactor # for the case Phi is not None, we can do much better than this
        
        if p == 2 and K.discriminant() % 2 == 0:
            prime_powers.append(base**ZZ(floor(4*8*log(2*d*a**2)/log(2)+2)))
        elif p == 3 and K.discriminant() % 3 == 0: # actually, we only need to do this if e(3) == 4 in the normal closure
            prime_powers.append(base**ZZ(floor(4*8*log(2*d*a**2)/log(3)+2)))
        else:
            prime_powers.append(base**ZZ(floor(4*log(2*d*a**2)/log(p)+1)))
    return prod(prime_powers)
            
        
    
    
    
